#!/usr/bin/env python

from __future__ import division, print_function, absolute_import

__author__ = "Marek Rudnicki"


import numpy as np

from . import spikes
from . import stats


def plot_neurogram(spike_trains, fs, ax=None, **kwargs):
    """Visualize `spike_trains` by converting them to bit map and plot
    using `plt.imshow()`.  Set `fs` reasonably in order to avoid
    aliasing effects.

    For smaller number of spike trains, it's usually better to use
    `plot_raster`.

    """

    neurogram = spikes.trains_to_array(
        spike_trains,
        fs
    )

    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()

    extent = (
        0,                       # left
        neurogram.shape[0] / fs, # right
        0,                       # bottom
        neurogram.shape[1]       # top
    )

    ax.imshow(
        neurogram.T,
        aspect='auto',
        extent=extent,
        **kwargs
    )

    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Channel number")

    return ax




def plot_raster(spike_trains, ax=None, style='k.', **kwargs):
    """Plot raster plot."""

    trains = spike_trains['spikes']
    duration = np.max( spike_trains['duration'] )

    # Compute trial number
    L = [ len(train) for train in trains ]
    r = np.arange(len(trains))
    n = np.repeat(r, L)

    # Spike timings
    s = np.concatenate(tuple(trains))


    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()

    ax.plot(s, n, style, **kwargs)
    ax.set_xlabel("Time (s)")
    ax.set_xlim( (0, duration) )
    ax.set_ylabel("Train Number")
    ax.set_ylim( (-0.5, len(trains)-0.5) )


    return ax





def plot_psth(
        spike_trains,
        bin_size,
        ax=None,
        drawstyle='steps-post',
        **kwargs
):
    """Plots PSTH of spike_trains."""


    psth, bin_edges = stats.psth(
        spike_trains,
        bin_size
    )


    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()


    ax.plot(
        bin_edges[:-1],
        psth,
        drawstyle=drawstyle,
        **kwargs
    )


    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Spikes per Second")


    return ax





def plot_isih(
        spike_trains,
        bin_size,
        ax=None,
        drawstyle='steps-post',
        density=True,
        **kwargs
):
    """Plot inter-spike interval histogram.

    """
    hist, bin_edges = stats.isih(
        spike_trains=spike_trains,
        bin_size=bin_size,
        density=density,
    )

    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()


    ax.plot(
        bin_edges[:-1],
        hist,
        drawstyle=drawstyle,
        **kwargs
    )


    ax.set_xlabel("Inter-Spike Interval (s)")

    if density:
        ax.set_ylabel("Probability Density Function")
    else:
        ax.set_ylabel("Interval Count per Bin")


    return ax






def plot_period_histogram(
        spike_trains,
        freq,
        nbins=64,
        shift=0,
        ax=None,
        style='',
        density=False,
        drawstyle='steps-post',
        **kwargs
):
    """Plot period histogram of the given spike trains.

    Parameters
    ----------
    spike_trains : spike_trains
        Spike trains for plotting.
    freq : float
        Stimulus frequency.
    nbins : int
        Number of bins for the histogram.
    shift : float
        Defines how much should the phase be shifted in the plot.
    ax : plt.Ax, optional
        Matplotlib Ax to plot on.
    style : str, optional
        Plotting style (See matplotlib plotting styles).
    density : bool, optional
        If False, the result will contain the number of samples in
        each bin. If True, the result is the value of the probability
        density function at the bin, normalized such that the integral
        over the range is 1. (See `np.histogram()` for reference)
    drawstyle : {'default', 'steps', 'steps-pre', 'steps-mid', 'steps-post'}
        Set the drawstyle of the plot.

    Returns
    -------
    plt.Axis
        Matplotlib axis containing the plot.

    """

    hist, bin_edges = stats.period_histogram(
        spike_trains,
        freq=freq,
        nbins=nbins,
        density=density
    )


    shift_samp = int(np.round( (shift*nbins) / (2*np.pi) ))
    hist = np.roll(hist, shift_samp)


    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca(polar=True)


    ax.plot(
        bin_edges[:-1],
        hist,
        style,
        drawstyle=drawstyle,
        **kwargs
    )

    # ax.plot(
    #     bin_edges[:-1],
    #     hist,
    #     style,
    #     **kwargs
    # )

    ax.set_xticks([0, np.pi, 2*np.pi])
    ax.set_xticklabels([0, r"$\pi$", r"2$\pi$"])

    ax.set_xlabel("Stimulus phase")

    if density:
        ax.set_ylabel("Probability density function")
    else:
        ax.set_ylabel("Spike count")

    return ax



def plot_sac(
        spike_trains,
        coincidence_window=50e-6,
        analysis_window=5e-3,
        normalize=True,
        ax=None,
        style='k-',
        **kwargs
):
    """Plot shuffled autocorrelogram (SAC) (Joris 2006)"""

    sac, bin_edges = stats.sac(
        spike_trains,
        coincidence_window=coincidence_window,
        analysis_window=analysis_window,
        normalize=normalize
    )


    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()


    ax.plot(
        bin_edges[:-1],
        sac,
        style,
        drawstyle='steps-post',
        **kwargs
    )


    ax.set_xlabel("Delay (s)")
    ax.set_ylabel("Normalized Number of Coincidences")

    return ax



def plot_signal(signal, fs=None, ax=None, style='', **kwargs):
    """Plot time signal.

    Parameters
    ----------
    signal : array_like
        Time signal.
    fs : float, optional
        Sampling freuency of the signal.
    ax : plt.Axis, optional
        Axis to plot onto.
    style : str, optional
        Plotting style string.

    Returns
    -------
    plt.Axis
       Matplotlib Axis with the plot.

    """

    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()


    if fs is None:
        fs = 1
        xlabel = "Sample Number"
    else:
        xlabel = "Time (s)"


    t = np.arange(len(signal)) / fs


    ax.set_xlim((t[0],t[-1]))

    ax.plot(t, signal, style, **kwargs)

    ax.set_xlabel(xlabel)
    ax.set_ylabel("Amplitude")

    return ax




def show():
    """Equivalent of plt.show()"""
    import matplotlib.pyplot as plt
    plt.show()


def gcf():
    """Equivalent of plt.gcf()"""
    import matplotlib.pyplot as plt
    fig = plt.gcf()
    return fig



#!/usr/bin/env python

from __future__ import division
__author__ = "Marek Rudnicki"

import numpy as np
from neuron import h
import waves as wv

def record_voltages(secs):
    vecs = []
    for sec in secs:
        vec = h.Vector()
        vec.record(sec(0.5)._ref_v)
        vecs.append(vec)
    return vecs


def plot_voltages(fs, vecs):
    import biggles

    all_values = np.concatenate( vecs )
    hi = all_values.max()
    lo = all_values.min()

    plot = biggles.Table(len(vecs), 1)
    plot.cellpadding = 0
    plot.cellspacing = 0
    for i,vec in enumerate(vecs):
        p = biggles.Plot()
        p.add( biggles.Curve(wv.t(fs, vec), vec) )
        p.yrange = (lo, hi)
        plot[i,0] = p


    p.add( biggles.LineX(0) )
    p.add( biggles.Label(0, (hi+lo)/2, "%.2f mV" % (hi-lo), halign='left') )

    p.add( biggles.LineY(lo) )
    p.add( biggles.Label((len(vec)/fs/2), lo, "%.1f ms" % (1000*len(vec)/fs), valign='bottom') )


    return plot


def main():
    pass

if __name__ == "__main__":
    import biggles
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This module implements permanent store for data.

"""
from __future__ import division, print_function, absolute_import
from __future__ import unicode_literals

__author__ = "Marek Rudnicki"

import os
import datetime
import logging
from itertools import izip_longest
import shelve
import collections

import tables

import numpy as np
import pandas as pd


logger = logging.getLogger('thorns')



def get_store(name='store', workdir='work'):
    """Return a quick and dirty shelve based persisten dict-like store.

    """

    fname = os.path.join(workdir, name + '.db')

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    store = shelve.open(fname, protocol=-1)

    return store




def dumpdb(data, name='dump', workdir='work', kwargs=None):
    """Dump data in order to recall the most up-to-date records later.

    Parameters
    ----------
    data : pd.DataFrame
        Data that will be appended to the database.
    name : str, optional
        Base name of the pickle file.
    workdir : str, optional
        Directory for the data.
    kwargs : dict, optional
        Additional parameters common for all data (MultiIndex will be
        extended).

    """
    fname = os.path.join(workdir, name+'.h5')

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    logger.info("Dumping data into {}.".format(fname))


    if kwargs is not None:
        for k,v in kwargs.items():
            data[k] = v
        data = data.set_index(kwargs.keys(), append=True)


    now = datetime.datetime.now()
    key = now.strftime("T%Y%m%d_%H%M%S_%f")

    store = pd.io.pytables.HDFStore(fname, 'a')
    store[key] = data

    store.close()





def loaddb(name='dump', workdir='work', timestamp=False, load_all=False):
    """Recall dumped data discarding duplicated records.

    Parameters
    ----------
    name : str, optional
        Base of the data filename.
    workdir : str, optional
        Directory where the data is stored.
    timestamp : bool, optional
        Add an extra column with timestamps to the index.
    load_all : bool, optional
        If True, data from all experiments will be loaded from the
        dumpdb file.  The default is to load only the most recent
        data.

    Returns
    -------
    pd.DataFrame
        Data without duplicates.

    """

    if timestamp:
        raise NotImplementedError("Should add extra columnt with timestamps to the index of the output.")


    fname = os.path.join(workdir, name+'.h5')
    store = pd.io.pytables.HDFStore(fname, 'r')


    logger.info("Loading data from {}".format(fname))

    if load_all:
        xkeys = collections.OrderedDict() # poor-man's ordered set
        dbs = []

        ### Get all tables from the store
        for t in sorted(store.keys()):
            df = store[t]

            # Just want ordered unique values in xkeys (ordered set would
            # simplify it: orderedset.update(df.index.names))
            for name in df.index.names:
                xkeys.setdefault(name)

            df = df.reset_index()
            dbs.append(df)


        db = pd.concat(dbs)


    else:
        last_key = sorted(store.keys())[-1]
        df = store[last_key]

        xkeys = df.index.names

        db = df.reset_index()


    store.close()



    ### Drop duplicates and set index
    db = db.drop_duplicates(
        subset=list(xkeys),
        take_last=True,
    )

    db = db.set_index(list(xkeys))



    return db



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Contains functions to read and write spike data in different
formats.

"""

from __future__ import print_function, division, absolute_import

__author__ = "Jörg Encke"

import struct
import pandas as pd
import copy
import warnings



def read_brainwaref32(filename, stimparams=None):
    """Read the spike timings as exported from BrainWare.

    BrainWare is software by Tucker-Davis Technologies.

    Parameters
    ----------
    filename : str
        The branwaref32 file to import.
    stimparams : dict
        A dict with the parameter names used in the stimulation
        sequence.  The key gives the parameter number as an integer
        while the value is the new name as a string.

    Returns
    -------
    spike_trains
        A DataFrame containing the spike timings in the thonrns
        spike_train format.

    """
    warnings.warn("read_brainwaref32: no unitest, might break in the future.")

    dict_list = []

    with open(filename,'rb') as f:

        while True:
            #read next 4 byte (32bit)
            s = f.read(4)
            if not s: break
            f32 = struct.unpack('f', s)

            if f32[0] == -2.0:
                #read number of parameters:
                s = f.read(4)
                if not s: break
                length = struct.unpack('f', s)

                s = f.read(4)
                if not s: break
                n_param = int(struct.unpack('f', s)[0])
                #read n parameter values
                param = struct.unpack('f' * n_param, f.read(n_param * 4))
                c_dict = {}

                #create a base dictonary with all parameters
                for i,v in enumerate(param):
                    name = "f%i" % i
                    c_dict[name] = v
                c_dict['spikes'] = []
                c_dict['duration'] = length[0] * 1E-3 #ms -> s

            elif f32[0] == -1.0:
                #c_dict has to be copied
                dict_list.append(copy.deepcopy(c_dict))

            else:
                if len(dict_list) > 0:
                    dict_list[-1]['spikes'].append(f32[0] * 1E-3) #ms -> s

    dataset = pd.DataFrame(dict_list)

    # Fill in column titles if given
    if stimparams is not None:
        for k,v in stimparams.iteritems():
            name = "f%i" % k
            dataset = dataset.rename(columns={name: v})


    return dataset



"""DSP related functions.

"""

from __future__ import division, print_function, absolute_import

__author__ = "Marek Rudnicki"


import numpy as np
import scipy.signal as dsp

from thorns.plotting import plot_signal, show


def align(a, fs_a, b, fs_b):
    """Align two signals, `a` and `b`, so that they have the same sampling
    frequency (resample to lower fs) and length (trim longer signal).

    """
    assert a.ndim == 1
    assert b.ndim == 1

    if fs_a == fs_b:
        fs = fs_a
    elif fs_a > fs_b:
        fs = fs_b
        a = resample(a, fs_a, fs)
    elif fs_a < fs_b:
        fs = fs_a,
        b = resample(b, fs_b, fs)

    if len(a) > len(b):
        a = a[0:len(b)]
    elif len(a) < len(b):
        b = b[0:len(a)]

    return a, b, fs


def snr(signal, noise):
    """Calculate signal-to-noise ratio in dB given `signal` and
    `noise`."""

    assert signal.shape == noise.shape

    snr_db = 20 * np.log10(
        rms(signal) / rms(noise)
    )

    return snr_db


def rms(signal):
    """Calculate root mean squere of a `signal`."""
    return np.sqrt(np.mean(signal**2))


def fft_filter(signal, fs, band):
    """Filter `signal` using a FFT filter.

    Parameters
    ----------
    signal : array_like
        Input signal.
    fs : float
        Sampling frequency of the input signal.
    band : tuple
        Tuple with lower and upler cut of frequencies: (lo, hi).


    Returns
    -------
    array_like
        Filtered signal.

    """

    lo, hi = band

    freqs = np.linspace(0, fs/2, len(signal)/2+1)

    signal_fft = np.fft.rfft(signal)
    signal_fft[(freqs < lo) | (freqs > hi)] = 0

    filtered = np.fft.irfft(signal_fft)

    return filtered


def set_dbspl(signal, dbspl):
    """Scale the level of `signal` to the given dB_SPL."""
    p0 = 20e-6
    rms = np.sqrt(np.sum(signal**2) / signal.size)

    scalled = signal * 10**(dbspl / 20.0) * p0 / rms

    return scalled


def resample(signal, fs, new_fs):
    """Resample `signal` from `fs` to `new_fs`."""
    new_signal = dsp.resample(signal, len(signal)*new_fs/fs)
    return new_signal


def trim(a, b):
    """Trim the longer vector, so that both have the same length."""
    assert a.ndim == b.ndim == 1

    length = min([len(a), len(b)])
    aa = a[0:length]
    bb = b[0:length]

    return aa, bb


def ramped_tone(
        fs,
        freq,
        duration,
        pad=0,
        pre=0,
        ramp=2.5e-3,
        dbspl=None
):
    """Generate ramped tone singal.

    Parameters
    ----------

    fs : float
        Sampling frequency in Hz.
    freq : float
        Frequency of the tone in Hz.
    duration : float
        Duration of the tone in seconds.
    pad : float, optional
        Duration of the pad in seconds (default is 0).  Pad will be
        appended at the end of the signal.
    pre : float, optional
        Duration of the pre-pad in seconds.  This pad will be attached
        at the end of the siganl.
    ramp : float, optional
        Duration of the ramp in seconds (default is 2.5 ms)
    dbspl : float, optional
        Amplitude of the tone in dB SPL.  If None (default), no
        scaling.  Scaling is done before ramping and appending the
        pad.


    Returns
    -------

    array_like
        The output tone with optional padding.

    """
    assert ramp < duration/2

    t = np.arange(0, duration, 1/fs)
    s = np.sin(2 * np.pi * t * freq)
    if dbspl is not None:
        s = set_dbspl(s, dbspl)

    if ramp != 0:
        ramp_signal = np.linspace(0, 1, np.ceil(ramp * fs))
        s[0:len(ramp_signal)] = s[0:len(ramp_signal)] * ramp_signal
        s[-len(ramp_signal):] = s[-len(ramp_signal):] * ramp_signal[::-1]

    pad_signal = np.zeros(pad * fs)
    pre_signal = np.zeros(pre * fs)
    sound = np.concatenate((pre_signal, s, pad_signal))

    return sound


def white_noise(
        fs,
        duration,
        band,
        seed,
        ramp=2.5e-3,
        pad=0,
        dbspl=None
):

    np.random.seed(seed)

    lo, hi = band

    n = int(np.round(duration * fs))
    freqs_abs = np.abs(np.fft.fftfreq(n, 1/fs))

    passband_mask = (freqs_abs > lo) & (freqs_abs < hi)

    angles = np.random.rand(n) * 2 * np.pi

    reals = np.cos(angles)
    imags = np.sin(angles)

    ss = reals + 1j*imags
    ss[np.invert(passband_mask)] = 0

    s = np.fft.ifft(ss).real

    s = set_dbspl(s, dbspl)

    # Ramping
    ramp_signal = np.linspace(0, 1, np.round(ramp*fs))
    s[0:len(ramp_signal)] = s[0:len(ramp_signal)] * ramp_signal
    s[-len(ramp_signal):] = s[-len(ramp_signal):] * ramp_signal[::-1]

    # Padding
    pad_signal = np.zeros(pad * fs)
    s = np.concatenate((s, pad_signal))

    return s


def electrical_pulse(
        fs,
        amplitudes,
        durations,
        gap=0,
        pad=0,
        charge=None
):
    """Generate electrical pulse.

    Parameters
    ----------
    fs : scalar
        Sampling frequency of the output.
    amplitudes : array_like
        A list of amplitudes in Ampere of each phase of the output
        pulse.
    durations : array_like
        A list of desired durations of each phase of the output pulse
        in seconds.  Must have the same number of elements as
        `amplitudes`.
    gap : scalar
        Desired duration of the gaps between phases of the output pulse.
    pad : scalar
        Desired duration of the zero signal appended at the end of the pulse.
    charge : None, scalar, optional
        If scalar, then it is equal to the absolute charge of the
        output pulse.  Amplitudes of the phases will be re-scaled, but
        will preserve their ratios to one other.

    Returns
    -------
    array_like
        Output signal (a pulse).

    """
    if not len(amplitudes) == len(durations):
        raise ValueError("`amplitudes` and `durations`" +
                         " must have the same length.")

    gap_signal = np.zeros(gap * fs)
    pad_signal = np.zeros(pad * fs)

    signals = []
    for amp, dur in zip(amplitudes, durations):
        signals.append(amp * np.ones(dur * fs))
        signals.append(gap_signal)

    # Remove the last gap
    signals.pop(-1)

    signals.append(pad_signal)

    signal = np.concatenate(signals)

    if charge is not None:
        charge_orig = np.sum(np.abs(
            np.array(amplitudes) * np.array(durations)
        ))
        signal *= charge / charge_orig

    return signal


def electrical_amplitudes(
        durations,
        polarity,
        ratio=None,
):
    """Calculate amplitudes for each "phase" of signle electrical pulse in
    cochlear implant.

    The resulting pulses are charged ballanced (if possible).  The
    function supports mono-, bi- and tri-phasic pulses.

    Parameters
    ----------
    durations : array_like
        List of phase durations in the pulse.
    polarity : {-1, 1, 'c' 'cathodic', 'a', 'anodic'}
        Polarity of the first phase.  -1, 'c' and 'cathodic' are
        equivalent.  As well as 1, 'a' and 'anodic'.
    ratio : float, optional
        Only valid for triphasic pulses.  It is equal to the charge
        ratio between the first and second phase.

    Returns
    -------
    tuple
        Amplitudes of phases corresponding to `durations`.

    """
    if len(durations) == 3:
        assert 0 <= ratio <= 1
    else:
        assert ratio is None

    # Normalize polarity
    if polarity in ('c', 'cathodic', -1):
        polarity = -1
    elif polarity in ('a', 'anodic', 1):
        polarity = 1
    else:
        raise RuntimeError("Unknown polarity")

    # Monophasic pulse
    if len(durations) == 1:
        amplitude = polarity / durations[0]
        amplitudes = (amplitude,)

    # Biphasic pulse
    elif len(durations) == 2:
        amplitudes = (
            +1*polarity * (0.5 / durations[0]),
            -1*polarity * (0.5 / durations[1])
        )

    # Triphasic pulse
    elif len(durations) == 3:
        amplitudes = (
            +1*polarity * (0.5 * ratio / durations[0]),
            -1*polarity * (0.5 / durations[1]),
            +1*polarity * (0.5 * (1-ratio) / durations[2])
        )

    else:
        raise RuntimeError("Unknown pulse shape")

    return amplitudes


def t(signal, fs):
    """Return time vector for `signal` with sampling frequency `fs` (Hz)."""

    return np.arange(0, len(signal)) / fs


def amplitude_modulated_tone(
        fs,
        fm,
        fc,
        m,
        duration,
        pad=0,
        ramp=2.5e-3,
        dbspl=None
):
    """Generate amplitude modulated tone.

    Parameters
    ----------
    fs : float
        Sampling frequency in Hz.
    fm : float
        Modulation frequency in Hz.
    fc : float
        Carrier frequency in Hz.
    m : float
        Modulation depth <0-1>.
    duration : float
        Tone duration in seconds.
    pad : float, optional
        Duration of the pad in seconds (default is 0)
    ramp : float, optional
        Duration of the ramp in seconds (default is 2.5 ms)
    dbspl : float, optional
        Amplitude of the tone in dB SPL.  If None (default), no scaling.


    Returns
    -------
    array_like
        AM signal.

    """
    assert ramp < duration/2
    assert 0 <= m <= 1
    assert fs/2 >= fc > fm

    t = np.arange(0, duration, 1/fs)
    s = (1 + m*np.sin(2*np.pi*fm*t)) * np.sin(2*np.pi*fc*t)

    if dbspl is not None:
        s = set_dbspl(s, dbspl)

    if ramp != 0:
        ramp_signal = np.linspace(0, 1, np.ceil(ramp * fs))
        s[0:len(ramp_signal)] = s[0:len(ramp_signal)] * ramp_signal
        s[-len(ramp_signal):] = s[-len(ramp_signal):] * ramp_signal[::-1]

    pad_signal = np.zeros(pad * fs)
    sound = np.concatenate((s, pad_signal))

    return sound


def roex(freqs, cf, p_lo, p_hi, w, t):
    """Calculate rounded exponential (roex) function.

    Currently the roex(p,w,t) version is implemented.

    """
    freqs_lo = freqs[freqs <= cf]
    freqs_hi = freqs[freqs > cf]

    g_lo = -(freqs_lo - cf) / cf
    g_hi = (freqs_hi - cf) / cf

    h_lo = (
        (1 - w)*(1 + p_lo*g_lo)*np.exp(-p_lo*g_lo) +
        w*(1 + p_lo*g_lo/t)*np.exp(-p_lo*g_lo/t)
    )

    h_hi = (
        (1 - w)*(1 + p_hi*g_hi)*np.exp(-p_hi*g_hi) +
        w*(1 + p_hi*g_hi/t)*np.exp(-p_hi*g_hi/t)
    )

    # w_hi = (1 + p_hi*g_hi)*np.exp(-p_hi*g_hi)

    h = np.concatenate((h_lo, h_hi))

    return h
